@@ -3,9 +3,12 @@ from datetime import datetime, timezone
 from typing import Union, Iterable
 
 
-def object_id_for_page(page):
-    # Predictable given that FPDF._putpages is invoked first in FPDF._enddoc:
-    return 2 * page + 1
+def buffer_subst(buffer, placeholder, value):
+    buffer_size = len(buffer)
+    assert len(placeholder) == len(value), f"placeholder={placeholder} value={value}"
+    buffer = buffer.replace(placeholder.encode(), value.encode(), 1)
+    assert len(buffer) == buffer_size
+    return buffer
 
 
 def format_date(date: datetime, with_tz=False) -> str:
